--[[
BagGnome, a World of Warcraft addon to manage item sets.
Copyright (C) 2007 Rabbit.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
]]

-------------------------------------------------------------------------------
-- Locals                                                                    --
-------------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("BagGnome")
local dewdrop = AceLibrary("Dewdrop-2.0")

-- GRRRR! We shouldn't need this variable, but BLIZZARD IS ON CRACK.
local createOrUpdateSetName = nil

local deequipQueue = nil
local queuedSet = {}
local slotLocks = nil
local queuedForDelete = nil
local activeSlots = nil
local missingItems = nil
local equipQueue = nil

local keybindingFrame = "BagGnomeKeybindingFrame"

local _G = getfenv(0)

local PickupContainerItem = _G.PickupContainerItem
local AutoEquipCursorItem = _G.AutoEquipCursorItem
local EquipCursorItem = _G.EquipCursorItem
local PickupInventoryItem = _G.PickupInventoryItem

-------------------------------------------------------------------------------
-- Addon declaration                                                         --
-------------------------------------------------------------------------------

BagGnome = AceLibrary("AceAddon-2.0"):new("AceHook-2.1", "AceConsole-2.0", "AceDB-2.0", "AceEvent-2.0")
local BagGnome = BagGnome

BagGnome.revision = tonumber(string.sub("$Revision: 68289 $", 12, -3))

local function IsEmpty()
	return next(BagGnome.OnMenuRequest.args.wear.args) == nil
end

local options = {
	type = "group",
	name = L["BagGnome"],
	desc = L["BagGnome options."],
	args = {
		add = {
			type = "execute",
			name = L["Add"],
			desc = L["Creates a new set, or updates an existing one."],
			func = "CreateOrUpdateSet",
			order = 101,
		},
		wear = {
			type = "group",
			name = L["Wear"],
			desc = L["Change equipment set."],
			pass = true,
			func = function(key)
				dewdrop:Close()
				if IsShiftKeyDown() then
					BagGnome:CreateOrUpdateSet(key)
				else
					BagGnome:WearSet(key)
				end
			end,
			args = {},
			order = 102,
			disabled = IsEmpty
		},
		key = {
			type = "group",
			name = L["Keybinding"],
			desc = L["Assign a keybinding to a set."],
			pass = true,
			get = function(key) return BagGnome.db.char.keybindings[key] end,
			set = function(key, value)
				if value == nil then
					BagGnome:ClearBinding(key)
				else
					BagGnome:RegisterBinding(value, key)
				end
			end,
			args = {},
			order = 103,
			disabled = IsEmpty
		},
		spacer = { type = "header", order = 105 },
		useBank = {
			type = "toggle",
			name = L["Use Bank"],
			desc = L["Use Bank to fill up bags and store non wanted bag items."],
			get = function() return BagGnome.db.profile.useBank end,
			set = function(v) BagGnome.db.profile.useBank = v end,
			order = 110,
		},		
		specialBags = {
			type = "toggle",
			name = L["Store Special Bags"],
			desc = L["Also store and restore special bags, like quivers and soul bags."],
			get = function() return BagGnome.db.profile.specialBags end,
			set = function(v) BagGnome.db.profile.specialBags = v end,
			order = 115,
		},				
		spacer = { type = "header", order = 120 },
		delete = {
			type = "group",
			name = L["Delete"],
			desc = L["Delete a set."],
			pass = true,
			func = "DeleteSet",
			args = {},
			order = 125,
			disabled = IsEmpty
		},
	},
}

-------------------------------------------------------------------------------
-- Initialization                                                            --
-------------------------------------------------------------------------------

function BagGnome:OnInitialize()
	self.version = (self.version or "1").."."..tostring(self.revision or "1")

	_G.BagGnome_VERSION = self.version
	_G.BagGnome_REVISION = self.revision

	self:RegisterDB("BagGnomeDB", "BagGnomePerCharDB")
	self:RegisterDefaults("profile", {
		specialBags = false,
		useBank = true,
	})
	self:RegisterDefaults("char", {
		set = {},
		icons = {},
		keybindings = {},
		alwaysWeapons = true,
	})

	self:RegisterChatCommand("/BagGnome", "/cg", options, "BAGGNOME")
	
	self.options = options
	
	-- XXX We're keeping this property for compability for a little while.
	self.OnMenuRequest = self.options
end

function BagGnome:OnEnable()
	if type(_G.StaticPopupDialogs) ~= "table" then
		_G.StaticPopupDialogs = {}
	end
	if not StaticPopupDialogs["BagGnomeAdd"] then
		StaticPopupDialogs["BagGnomeAdd"] = {
			text = L["|cffeda55fLeft-Click|r a slot to toggle it for this set. Green slots are |cff00ff00enabled|r, red are |cffff0000disabled|r, yellow are |cffffff00missing items|r.\n\nIf you want to assign an |cff0000fficon|r to this set, you can |cffeda55fCtrl-Click|r a slot to use that slots item as the icon.\n\nType the desired set name below and click Add when you are done."],
			button1 = L["Add"],
			button2 = L["Cancel"],
			OnCancel = function() self:CancelCreateSet() end,
			sound = "levelup2",
			whileDead = 1,
			hideOnEscape = 1,
			timeout = 0,
			OnShow = function()
				-- We have to do this onshow to reset the previous text
				_G[this:GetName().."EditBox"]:SetText(createOrUpdateSetName or "")
			end,
			OnHide = function()
				_G[this:GetName().."EditBox"]:SetText("")
			end,
			EditBoxOnEnterPressed = function()
				local name = _G[this:GetParent():GetName().."EditBox"]:GetText()
				self:AddSetFromDoll(name)
				createOrUpdateSetName = nil
				this:GetParent():Hide()
			end,
			EditBoxOnEscapePressed = function()
				self:CancelCreateSet()
				this:GetParent():Hide()
			end,
			OnAccept = function()
				local name = _G[this:GetParent():GetName().."EditBox"]:GetText()
				self:AddSetFromDoll(name)
				createOrUpdateSetName = nil
			end,
			hasEditBox = 1,
		}
	end
	if not StaticPopupDialogs["BagGnomeDelete"] then
		StaticPopupDialogs["BagGnomeDelete"] = {
			text = L["Are you sure you want to delete the set %s?"],
			button1 = L["Delete"],
			button2 = L["Cancel"],
			showAlert = 1,
			timeout = 0,
			OnAccept = function() self:DeleteSet(queuedForDelete, true) end,
			OnCancel = function() queuedForDelete = nil end,
		}
	end

	-- Create the button frames for keybindings
	for k, v in pairs(self.db.char.keybindings) do
		self:RegisterBinding(v, k)
	end
	
	self:RegisterEvent("BANKFRAME_OPENED", 				function() self.bankOpened = true  end)
	self:RegisterEvent("BANKFRAME_CLOSED", 				function() self.bankOpened = false end)		
	self:UpdateSetOptions()
end

local function reclaim(t)
	if type(t) ~= "table" then return end
	for k, v in pairs(t) do
		if type(v) == "table" then
			t[k] = reclaim(v)
		else
			t[k] = nil
		end
	end
	t = nil
	return nil
end

local function LockSlot(bag, slot)
	if not slotLocks then slotLocks = {} end
	if not slotLocks[bag] then slotLocks[bag] = {} end
	slotLocks[bag][slot] = true
end

local function IsSlotLocked(bag, slot)
	if slotLocks and slotLocks[bag] and slotLocks[bag][slot] then
		return true
	end
	return false
end

local function ClearSlotLocks()
	slotLocks = reclaim(slotLocks)
end

function BagGnome:OnDisable()
	activeSlots = reclaim(activeSlots)
	missingItems = reclaim(missingItems)
	ClearSlotLocks()

	for k, v in pairs(self.db.char.keybindings) do
		self:UnsetBinding(k)
	end
end


-------------------------------------------------------------------------------
-- Keybindings                                                               --
-------------------------------------------------------------------------------

local function ValidateBinding(combo)
	return AceLibrary("AceConsole-2.0").keybindingValidateFunc(combo)
end

local function KeybindingClicked()
	if type(this.set) ~= "string" or not ValidateBinding(this.combo) then return end
	BagGnome:WearSet(this.set)
end

function BagGnome:RegisterBinding(combo, set)
	if not combo or not set then return end

	if not ValidateBinding(combo) then
		self:Print(L["%s is not a valid keybinding."]:format("|cffd9d919"..combo.."|r"))
		return
	end

	local shouldPrint = true
	for k, v in pairs(self.db.char.keybindings) do
		if v == combo then
			shouldPrint = nil
		elseif k == set then
			self:UnsetBinding(set)
		end
	end

	if shouldPrint then
		self:Print(L["Registering keybinding %s to set %s."]:format("|cffd9d919"..combo.."|r", "|cffd9d919"..set.."|r"))
	end

	local frameName = keybindingFrame..combo
	local frame = _G[frameName] or CreateFrame("Button", frameName, _G["UIParent"])
	frame.combo = combo
	frame.set = set
	frame:SetScript("OnMouseUp", KeybindingClicked)

	self.db.char.keybindings[set] = combo
	SetBindingClick(combo, frameName)

	self:TriggerEvent("BagGnome_KeybindingsUpdated")
end

function BagGnome:UnsetBinding(set)
	if not set or not self.db.char.keybindings[set] then return end
	local combo = self.db.char.keybindings[set]

	local frame = _G[keybindingFrame..combo]
	if frame then
		frame:SetScript("OnMouseUp", nil)
		frame = nil
	end

	SetBinding(combo) -- Unset the binding.

	self:TriggerEvent("BagGnome_KeybindingsUpdated")
end

function BagGnome:ClearBinding(set)
	if not set or not self.db.char.keybindings[set] then return end

	self:UnsetBinding(set)

	self:Print(L["Removing keybinding %s from set %s."]:format("|cffd9d919"..self.db.char.keybindings[set].."|r", "|cffd9d919"..set.."|r"))
	self.db.char.keybindings[set] = nil

	self:TriggerEvent("BagGnome_KeybindingsUpdated")
end

-------------------------------------------------------------------------------
-- Menu                                                                      --
-------------------------------------------------------------------------------

function BagGnome:UpdateSetOptions()
	local opt = self.OnMenuRequest.args

	opt.wear.args = reclaim(opt.wear.args)
	opt.delete.args = reclaim(opt.delete.args)
	opt.key.args = reclaim(opt.key.args)

	opt.wear.args = {}
	opt.delete.args = {}
	opt.key.args = {}

	for k, v in pairs(self.db.char.set) do
		local icon = self.db.char.icons[k]
		if not self:HideSetFromUI(k) then
			opt.wear.args[k] = {
				type = "execute",
				name = k,
				desc = L["Equip %s, or hold down Shift to edit this set."]:format(k),
				icon = icon,
			}
		end
		opt.delete.args[k] = {
			type = "execute",
			name = k,
			desc = L["Deletes the equipment set %s."]:format(k),
			icon = icon,
		}
		opt.key.args[k] = {
			type = "text",
			name = k,
			desc = L["Assign a keybinding to %s (or empty to clear)."]:format(k),
			validate = "keybinding",
			icon = icon,
		}
	end
end

-------------------------------------------------------------------------------
-- Addon methods                                                             --
-------------------------------------------------------------------------------

local currentlyMouseOveredSlot = nil
do
	local green = L["|cff00ff00Green slots|r are active, and any item currently in a green slot will be saved as part of the item set."]
	local red = L["|cffff0000Red slots|r are disabled, and will be ignored by the item set."]
	local yellow = L["A |cffffff00yellow slot|r means the item is missing from your inventory, and when you update the set, the item stored in the set will be used, and not the currently equipped one."]
	local blue = L["A |cff0000ffblue slot|r tells your BagGnome to use the item in that slot as the set icon. If a slot is toggled to an icon slot, it is also activated immediately."]
	function BagGnome:OnTooltipSetItem(tooltip, ...)
		if not activeSlots then return end
		if activeSlots.icon and activeSlots.icon == currentlyMouseOveredSlot then
			tooltip:AddLine(blue, 0.35, 0.6, 0.95, true)
		elseif activeSlots[currentlyMouseOveredSlot] then
			tooltip:AddLine(green, 0.35, 0.6, 0.95, true)
		elseif missingItems and missingItems[currentlyMouseOveredSlot] then
			tooltip:AddLine(yellow, 0.35, 0.6, 0.95, true)
		elseif type(activeSlots[currentlyMouseOveredSlot]) == "boolean" and not activeSlots[currentlyMouseOveredSlot] then
			tooltip:AddLine(red, 0.35, 0.6, 0.95, true)
		end
		return self.hooks[tooltip].OnTooltipSetItem(tooltip, ...)
	end
end

local function OnEnter(...)
	currentlyMouseOveredSlot = this.id
	if type(PaperDollItemSlotButton_OnEnter) == "function" then
		this = this:GetParent()
		PaperDollItemSlotButton_OnEnter(this, ...)
	end
end

local function OnLeave(...)
	currentlyMouseOveredSlot = nil
	this:GetParent().updateTooltip = nil
	GameTooltip:Hide()
	ResetCursor()
end

local function OnClick(...)
	local id = this.id

	if IsShiftKeyDown() or (not IsModifierKeyDown() and CursorHasItem()) and type(PaperDollItemSlotButton_OnClick) == "function" then
		this = this:GetParent()
		PaperDollItemSlotButton_OnClick("LeftButton")
	elseif IsControlKeyDown() then
		-- Don't want an empty slot as icon.
		local currentItem = BagGnome:ItemNameFromSlot(id)
		if currentItem == false then return end

		if activeSlots["icon"] then
			local oldSlot = activeSlots["icon"]
			activeSlots[oldSlot] = true
			_G["BagGnomeTexture"..oldSlot]:SetVertexColor(0, 255, 0)
		end
		activeSlots["icon"] = id
		activeSlots[id] = true
		this.texture:SetVertexColor(0, 0, 255)
	else
		activeSlots[id] = not activeSlots[id]
		if activeSlots[id] then
			if activeSlots["icon"] == id then
				this.texture:SetVertexColor(0, 0, 255)
			else
				missingItems[id] = nil
				this.texture:SetVertexColor(0, 255, 0)
			end
		else
			this.texture:SetVertexColor(255, 0, 0)
		end
	end
end

-- Opens the character frame and lets the user uncheck the items he doesn't
-- want.
function BagGnome:CreateOrUpdateSet(input)
	if UnitAffectingCombat("player") or InCombatLockdown() then return end

	OpenAllBags(1)

--	if not self:IsHooked("CharacterFrame_OnHide") then
	--	self:Hook("CharacterFrame_OnHide", true)
--	end
	-- hook inventory on hide
	if not self:IsHooked(GameTooltip, "OnTooltipSetItem") then
		self:HookScript(GameTooltip, "OnTooltipSetItem")
	end

	dewdrop:Close()

	createOrUpdateSetName = input
	StaticPopup_Show("BagGnomeAdd")
end

function BagGnome:CancelCreateSet()
	self:CharacterFrame_OnHide()
end

function BagGnome:AddSetFromDoll(name)
	self:AddSet(name, slots, icon)
	self:CharacterFrame_OnHide()
end

function BagGnome:AddSet(name, slots, iconSlot)
	local concatedName = name:gsub(" ", "")
	if not name or name == "" or concatedName == "" then
		self:Print(L["Please use a proper name for your set."])
		return
	end
	if self:HasSet(concatedName) then name = concatedName end

	local text = L["Added set: %s."]:format(name)
	local event = "BagGnome_AddSet"
	if self:HasSet(name) then
		text = L["Updating set: %s."]:format(name)
		event = "BagGnome_UpdateSet"
		reclaim(self.db.char.set[name]);
	end
	
	self.db.char.set[name] = {};
	
	--self:Print("adding bag data");

	-- add bag data here
	for bag = 0, 4 do		
		if (self.db.profile.specialBags) or (self:IsNormalBag(bag)) then
			local size = GetContainerNumSlots(bag) or 0
			self.db.char.set[name][bag] = {};
			for slot = 1, size do
		  	local link = GetContainerItemLink(bag, slot)
		  	local itemName = BagGnome:GetItemName(link);
		  	local _, qty = GetContainerItemInfo(bag, slot)
	  		--local _, _, code = strfind(link or "", "(item:%d+:%d+:%d+:%d+:%d+:%d+:%-?%d+:%-?%d+)")
	  		--local info = code and (code .. "," .. qty) or nil
		  	if (itemName) then
		  		--self:Print(bag.."-"..slot.."- "..itemName.." | "..qty);
	  		end
	  		self.db.char.set[name][bag][slot] = {};
	  		self.db.char.set[name][bag][slot].itemName = itemName;
	  		self.db.char.set[name][bag][slot].qty = qty;	  	
	  	end
		end
	end

	if not self.db.char.set[name] then return end

	if iconSlot then
		self.db.char.icons[name] = GetInventoryItemTexture("player", iconSlot)
	end

	self:TriggerEvent(event, name)
	self:Print(text)

	self:UpdateSetOptions()
end

function BagGnome:DeleteSet(name, verified)
	if not verified and not IsShiftKeyDown() then
		queuedForDelete = name
		StaticPopup_Show("BagGnomeDelete", name)
		return
	end

	queuedForDelete = nil
	if self.db.char.set[name] then
		self.db.char.set[name] = reclaim(self.db.char.set[name])
		self.db.char.icons[name] = nil
		self:ClearBinding(name)

		self:TriggerEvent("BagGnome_DeleteSet", name)
		self:Print(L["Deleted set: %s."]:format(name))

		self:UpdateSetOptions()
		dewdrop:Refresh()
	end
end

-- ERR_2HANDED_EQUIPPED

function BagGnome:WearSet(...)
	ClearCursor()
	if CursorHasItem() or CursorHasSpell() or (UnitIsDeadOrGhost("player") and not UnitIsFeignDeath("player")) then return false end

	local size = select("#", ...)
	if size < 1 then return false end

	local setToWear = nil
	local name = nil

	if size == 1 then
		name = select(1, ...)
		setToWear = self.db.char.set[name]
	else
		for i = 1, size do
			local set = select(i, ...)
			if type(set) ~= "string" or not self.db.char.set[set] then return false end
		end
		setToWear = {}
		for i = 1, size do
			name = select(i, ...)
			for k, v in pairs(self.db.char.set[name]) do
				setToWear[k] = v
			end
		end
	end
	if type(setToWear) ~= "table" then return false end

	if UnitAffectingCombat("player") then
		if size > 1 then
			setToWear = reclaim(setToWear)
		end
		return false
	end

	self:Print(L["Restoring your inventory now!"])	
	if (self.bankOpened) then
		self:Print(L["Using Bank to fill up stacks !"])	
	end
	self.presentBag = 0;
	self.presentSlot = 1;
	self.presentSetName = name;
	self:UnPause();
	
	self:RegisterEvent("BAG_UPDATE");
	
	self:BagMoveFunction(name);	
	
	self:TriggerEvent("BagGnome_WearSet", name)	

	if size > 1 then
		setToWear = reclaim(setToWear)
	end

	return true
end

function BagGnome:Pause(bag,slot,itemName,qty) -- set fillbag to true if you are calling inside the PlaceItemInSlot function	
	self.paused = 1;	
	self.fullfillBag = bag;
	self.fullfillSlot = slot;
	self.fullfillItem = itemName;
	self.fullfillQty = qty;				
	self:ScheduleRepeatingEvent("BagGnome_BagMove",self.BagMoveFunction, 0.2, self);
	self.bagUpdateCount = 0;
	--self:Print("Paused !");
end

function BagGnome:UnPause()
	self.paused = nil;
	self.fullfillBag = nil;
	self.fullfillSlot = nil;
	self.fullfillItem = nil;
	self.fullfillQty = nil;					
	self:CancelScheduledEvent("BagGnome_BagMove");
	--self:Print("Unpaused !");
	self:BagMoveFunction();
end

function BagGnome:BAG_UPDATE()
	--self:Print("BagUpdate ! Start");
	self.bagUpdateCount = self.bagUpdateCount+1;
	if (self.bagUpdateCount > 10) then -- abort if there is a problem
		self:Print("Error : Sorting failed ! Internal problem ! Restart sorting maybe it works now !");
		self:CancelScheduledEvent("BagGnome_BagMove");
		self:UnregisterEvent("BAG_UPDATE");
		self.paused = nil;
		return;
	end
	if (self.fullfillBag) then
		--self:Print("Looking in : "..self.fullfillBag.." - "..self.fullfillSlot);
		local link = GetContainerItemLink(self.fullfillBag, self.fullfillSlot)
		local itemName = BagGnome:GetItemName(link);
		local _, qty = GetContainerItemInfo(self.fullfillBag, self.fullfillSlot)
		--self:Print("BagUpdate Browsing !");
		if ( (itemName == self.fullfillItem) and (qty == self.fullfillQty) ) then -- moving completed, lets unpause
			self:UnPause();			
			return;
		end		
	end	
end

function BagGnome:BagMoveFunction()
	--self:Print("BagMoveCheck !");
	if (self.paused) then 
		self:BAG_UPDATE();
		return false;
	end
	
	local name = self.presentSetName;
	local startBag = self.presentBag;
	
	for bag = self.presentBag, 4 do		
		self.presentBag = bag
		local size = GetContainerNumSlots(bag) or 0
		if (self.db.char.set[name][bag]) then
			local start = 1;
			if (bag == startBag) then start = self.presentSlot; end
			--self:Print("Resuming from slot : "..start);
			for slot = start, size do
				self.presentSlot = slot
				if (self.db.char.set[name][bag][slot]) then
			  	local link = GetContainerItemLink(bag, slot)
			  	local itemName = BagGnome:GetItemName(link);
		  		local _, qty = GetContainerItemInfo(bag, slot)
		  		if (itemName) then -- item actually in that bag spot
		  			if (self.db.char.set[name][bag][slot].itemName) then -- somethings should be there, see if its right
			  			if (self.db.char.set[name][bag][slot].itemName == itemName) then -- correct item is there
			  				--self:Print("Same item : "..bag.."-"..slot.."- "..itemName.." | "..qty);
			  				if (self.db.char.set[name][bag][slot].qty == qty) then -- even same qty = do nothing cos its good
		  						--self:Print("Same qty as well !");
		  					else
		  						--self:Print("But qty mismatch !"..qty.." avail/ "..self.db.char.set[name][bag][slot].qty.." wanted"); -- more or less ?		  						
		  						if (qty < self.db.char.set[name][bag][slot].qty) then -- qty less than waht should be, order rest
		  							local newQty = self.db.char.set[name][bag][slot].qty - qty;
		  							self:PlaceItemInSlot(bag,slot,self.db.char.set[name][bag][slot].itemName,newQty,qty); -- request missing, qty is alreayd there
		  							return false;
		  						else -- more than it should be , put some away
		  							local remQty = qty - self.db.char.set[name][bag][slot].qty;
										local toBag, toSlot = self:LocateFreeSlot();
										if toBag ~= nil then -- slot found to move to
											--self:Print("Moving Part of Stack to : "..toBag.."-"..toSlot.."- "..itemName.." | "..remQty);																						
											self:Pause(bag,slot,self.db.char.set[name][bag][slot].itemName,self.db.char.set[name][bag][slot].qty); -- that what should be, should remain there
											SplitContainerItem(bag,slot,remQty);
											PickupContainerItem(toBag,toSlot);																						
											return false;
										end		  							
										self:Print("Abort : Could not free space for item move !");
										return false;
		  						end		  						
		  					end
		  				else -- wrong item there remove , place new one
		  					--self:Print("Wrong item found : "..bag.."-"..slot.."- "..itemName.." | "..qty);		  					  		
		  					self:RemoveItemFromSlot(bag,slot,itemName,qty);		  				-- just remove it and next pass will place the correct one there
		  					return false;
		  				end
		  			else -- item there but shouldnt !
		  				--self:Print("Found item which shouldnt be there : "..bag.."-"..slot.."- "..itemName.." | "..qty);		  					  		
		  				self:RemoveItemFromSlot(bag,slot,itemName,qty);		  				
		  				return false;
		  			end
		  		else -- nothing there, lets see if some should have been there		  			
		  			if (self.db.char.set[name][bag][slot].itemName) then -- something should have been there
		  				--self:Print("Nothing in slot : "..bag.."-"..slot.." but should be - "..self.db.char.set[name][bag][slot].itemName.." | "..self.db.char.set[name][bag][slot].qty);
		  				self:PlaceItemInSlot(bag,slot,self.db.char.set[name][bag][slot].itemName,self.db.char.set[name][bag][slot].qty,0);		  				
		  				return false;
		  			else -- nothing should be there, correct
		  				--self:Print("Slot correctly empty : "..bag.."-"..slot);		  					  		
		  			end		  					  					  			
		  		end		  		
		  	else		  		
		  		self:Print("Error : In Bag "..bag.." -Slot .."..slot.." not stored in set "..name);
		  	end
		  end
		else
			if (self:IsNormalBag(bag)) then
				self:Print("Error : Bag "..bag.." not stored in set "..name);
			end
		end
	end	
	if (self:IsEventRegistered("BAG_UPDATE")) then
		self:UnregisterEvent("BAG_UPDATE");
	end
	self:Print(L["Inventory restore finished !"])		
end

function BagGnome:RemoveItemFromSlot(bag,slot,itemName,qty)
	--self:Print("Removing Item from : "..bag.."-"..slot);		  					  		
	local toBag, toSlot = self:LocateFreeSlot();
	if toBag ~= nil then -- slot found to move to
		--self:Print("RemoveItem - Found space at : "..toBag.."-"..toSlot);		  					  		
		self:Pause(toBag,toSlot,itemName,qty);
		PickupContainerItem(bag,slot);
		PickupContainerItem(toBag,toSlot);						
		return true;		
	end
	self:Print("Abort : Could not free space for item move !");
	return false;
end


function BagGnome:PlaceItemInSlot(bag,slot,itemName,wQty,avail) -- wqty - what needs to be put there to fullfill requirements / avail - what is already in there
	--self:Print("Requested item : "..bag.."-"..slot.."- "..itemName.." | "..wQty);		  					  		
	local bagNum, slotNum = self:FindItem(itemName);
	if (bagNum > -1) then -- item found		
		local _, presentqty = GetContainerItemInfo(bagNum, slotNum)								
		--self:Print("Found Item in : "..bagNum.."-"..slotNum.."- "..itemName.." | "..presentqty);		  					  		
		local newQty;
		if (presentqty >= wQty) then
			--self:Print("Fully fullfilling request : "..bagNum.."-"..slotNum.."- "..itemName.." | "..wQty);		  					  		
			SplitContainerItem(bagNum,slotNum,wQty);
			newQty = wQty + avail; -- that will be in there after move
		else 
			--self:Print("Partly fullfilling request : "..bagNum.."-"..slotNum.."- "..itemName.." | "..presentqty);		  					  						
			PickupContainerItem(bagNum,slotNum);
			newQty = presentqty + avail; -- that will be in there after move
		end		
		self:Pause(bag,slot,itemName,newQty);
		PickupContainerItem(bag,slot);				
	else
		self:Print("Item : "..itemName.." not enough available "..bag.."/"..slot.." - "..avail.."/"..wQty);
		self.presentSlot = self.presentSlot + 1; -- continue with next slot
		self:BagMoveFunction();
		return false;			
	end
	return true;
end

function BagGnome:GetItemName(itemLink)
	if (not itemLink) then return nil; end
	assert(itemLink);
	local itemName = GetItemInfo(itemLink);
	return itemName;
end

function BagGnome:FindItem(item)
	assert(item);
	for i = self.presentBag, 4, 1 do -- locate stuff in inv first then bank
		local start = 1;
		if (i == self.presentBag) then
			start = self.presentSlot+1;
		end
		for j = start,GetContainerNumSlots(i),1 do			
			local link = GetContainerItemLink(i, j)			
			local itemName = BagGnome:GetItemName(link);
			--if (link) then
				--self:Print("Browsing for item : "..item.." | "..i.." - "..j.." | "..itemName);		  					  		
			--end			
			if itemName and item == itemName then return i, j end
		end
	end
	if ( (self.bankOpened) and (self.db.profile.useBank) ) then
		for i = 5, 11, 1 do -- locate stuff in bank
			for j = 1,GetContainerNumSlots(i),1 do			
				local link = GetContainerItemLink(i, j)
				local itemName = BagGnome:GetItemName(link);
				--if (link) then
					--self:Print("Browsing for item : "..item.." | "..i.." - "..j.." | "..itemName);		  					  		
				--end			
				if itemName and item == itemName then return i, j end
			end
		end
	end
	return -1
end

function BagGnome:LocateFreeSlot()
	if ( (self.bankOpened) and (self.db.profile.useBank) ) then -- find free spot in bank first
		for theBag = 11, 5, -1 do
			if (self:IsNormalBag(theBag)) then
				local numSlot = GetContainerNumSlots(theBag)
				for theSlot = numSlot, 1, -1 do
					local texture = GetContainerItemInfo(theBag, theSlot)
					if not texture then
						return theBag, theSlot
					end
					end
			end
		end
	end
	for theBag = 4, 0, -1 do
		if (self:IsNormalBag(theBag)) then
			local numSlot = GetContainerNumSlots(theBag)
			for theSlot = numSlot, 1, -1 do
				local texture = GetContainerItemInfo(theBag, theSlot)
				if not texture then
					return theBag, theSlot
				end
			end
		end
	end
	return nil
end

-------------------------------------------------------------------------------
-- Utility                                                                   --
-------------------------------------------------------------------------------

function BagGnome:GetItemSubtype(link)
	return select(7, GetItemInfo(link))
end

function BagGnome:IsNormalBag(bagId)
	if bagId == 0 or bagId == -1 then return true end
	local link = GetInventoryItemLink("player", ContainerIDToInventoryID(bagId))
	if not link then return false end
	local linkId = select(3, link:find("item:(%d+)"))
	if not linkId then return false end
	local bagType = self:GetItemSubtype(linkId)
	if bagType and bagType == L["Bag"] then return true end
	return false
end

function BagGnome:HasSet(name)
	return self.db.char.set[name] ~= nil
end

-------------------------------------------------------------------------------
-- Hooks                                                                     --
-------------------------------------------------------------------------------

-- Only hooked if we are showing the paper doll frame with set adding
function BagGnome:CharacterFrame_OnHide()
	StaticPopup_Hide("BagGnomeAdd")

	self:UnhookAll()

	CloseAllBags()	
end

function BagGnome:Noop() --[[ noop ]] end

-------------------------------------------------------------------------------
-- Events                                                                    --
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
-- UI                                                                        --
-------------------------------------------------------------------------------

-- Hook this method and return true for any sets you do not want listed in the
-- tablet.
function BagGnome:HideSetFromUI(set)
	return false
end

-- Hook to provide for example coloring of different sets in the tablet.
function BagGnome:GetSetDisplayName(set)
	return set
end
